/*
multimux  copyright 2003 Jan Panteltje.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "multimux.h"


char *strsave(char *s) /* save string s somewhere */
{
char *p;

p = (char *) malloc(strlen(s) + 1);
if(!p) return 0;

strcpy(p, s);
return p;
} /* end function strsave */


int print_usage()
{
fprintf(stderr, "\
Usage:               multimux [-d] [-h] -o outputfile [-w] infile1 infile2 ... infile7\n\n\
-d delay2,...delayn  delays in milliseconds for audio channels relative to first channel\n\
                      (the first channel must be the earliest)\n\
-h                   help (this help)\n\
-o outputfile        multiplexed .wav output filename\n\
-v                   debug (prints functions and arguments)\n\
      (use for multi channel ac3: ffmpeg -i 4ch.wav -ab 256 -ac 4 -acodec ac3 4ch.ac3\n\
");

return 1;
} /* end function print_usage */


int main(int argc, char **argv)
{
int a, i, j;
double da, db;
int header_size;
int multimux_header_size;
wave_header *out_header;
wave_header *in_header[MAX_FILES];
FILE *ifptr[MAX_FILES];
FILE *ofptr;
char *input_filename[MAX_FILES];
char *output_filename;
int input_files;
struct stat *statptr[MAX_FILES];
int skip_argument_flag;
unsigned char uc[10]; /* max number of bytes per sample, using only 2 here */
double delay[MAX_FILES];
char *delay_string;
char *token;
const char delimiters[] = ",";
size_t bytes_to_add[MAX_FILES];
size_t bytes_added[MAX_FILES];
size_t longest_input_file_length;
size_t longest_input_data_length;
size_t expected_output_filesize;
size_t bytes_written;
size_t bytes_read[MAX_FILES];
int display_counter;
size_t sta;

fprintf(stderr, "Panteltje multimux-%s \n", VERSION);

/* defaults */
output_filename = 0;
multimux_header_size = 0;
delay_string = 0;
longest_input_data_length = 0;

for(i = 0; i < MAX_FILES; i++)
	{
	delay[i] = 0.0;
	}

/* end defaults */

while(1)
	{
	a = getopt(argc, argv, "d:ho:v");
	if(a == -1) break;

	switch(a)
		{
		case 'd':
			delay_string = strsave(optarg);
			if(! delay_string)
				{
				fprintf(stderr, "Could not allocate space for delay_string, abnorting.\n");

				exit(1);
				}

			break;
		case 'h':
			print_usage();
			exit(1);
			break;
		case 'o':
			output_filename = strsave(optarg);
			if(! output_filename)
				{
				fprintf(stderr, "Could not allocate space for output_filename, aborting.\n");

				exit(1);
				}

			break;
		case 'v':
			debug_flag = 1;
			break;
        case -1:
        	break;
		case '?':
			if (isprint(optopt) )
 				{
 				fprintf(stderr, "Unknown option `-%c'.\n", optopt);
 				}
			else
				{
				fprintf(stderr, "Unknown option character `\\x%x'.\n",\
				optopt);
				}
			print_usage();

			exit(1);
			break;			
		default:
			print_usage();

			exit(1);
			break;
		}/* end switch a */
	}/* end while getopt() */

if(debug_flag)
	{
	// are we on a 16, 32, or 64 bit system?
	fprintf(stderr, "unsigned int max=%u unsigned long max=%lu\n", UINT_MAX, ULONG_MAX);
	}

if(! output_filename)
	{
	fprintf(stderr, "No output filename specified (-o flag), aborting (multimux -h for help).\n");
	print_usage();

	exit(1);
	}

skip_argument_flag = 0;
input_files = 0;
for(i = 0; i < argc - 1; i++)
	{
	if(argv[i + 1][0] == '-')
		{
		skip_argument_flag = 1;
		continue;
		}

	if(! skip_argument_flag)
		{
		input_filename[input_files] = strsave(argv[1 + i]);
		if(! input_filename[input_files])
			{
			fprintf(stderr, "Could mot allocate space for input_filename[%d], aborting.\n", input_files);

			exit(1);
			}
 
		if(debug_flag)
			{
			fprintf(stderr, "found input file %s\n", input_filename[input_files]);
			}

		input_files++;
		} /* end if ! command line flag */

	skip_argument_flag = 0;
	} /* end while al input files */

if(input_files == 0)
	{
	fprintf(stderr, "No input file(s) specified, aborting (multimux -h for help).\n");

	exit(1);
	}

if(input_files == 1)
	{
	fprintf(stderr, "WARNING only 1 input file specified!\n");
	}

/* get size of output wave header */
header_size = sizeof(wave_header);

if(debug_flag)
	{
	fprintf(stderr, "main(): wave header_size=%d\n", header_size);
	}

/* allocate space for output wave header */
out_header = (wave_header *) malloc( sizeof(wave_header) );
if(! out_header)
	{
	fprintf(stderr,\
	"multimux(): could not allocate space for out_header, aborting.\n");

	exit(1);
	}

/* try to open the output file */
ofptr = fopen(output_filename, "w");
if(! ofptr)
	{
	fprintf(stderr, "Could not open output file %s for write, aborting.\n", output_filename);

	exit(1);
	}
	
/* try to open the input file(s) */
for(i = 0; i < input_files; i++)
	{
	ifptr[i] = fopen(input_filename[i], "r");

	if(! ifptr[i])
		{
		fprintf(stderr, "Could not open file %s for read, aborting.\n", input_filename[i]);

 		exit(1);
		}

	bytes_read[i] = 0;	
	} /* end for all possible input files */

if(debug_flag)
	{
	fprintf(stderr, "input_files=%d\n", input_files);
	}

/* parse delay string */
for(i = 0; i < input_files; i++)
	{
	if(i == 0) token = strtok (delay_string, delimiters);
	else token = strtok (NULL, delimiters);
	if(! token) break;

	sscanf(token, "%lf", &delay[i]);

	if(debug_flag)
		{
		fprintf(stderr, "delay[%d]=%.2f\n", i, delay[i]);
		}

	} /* end for all delays (as many delays as input files) */

longest_input_file_length = 0;
for(i = 0; i < input_files; i++)
	{
	/* allocate space for structure stat so we can get length input file */
	statptr[i] = (struct stat*) malloc( sizeof(struct stat) );
    if(! statptr[i])
		{
		fprintf(stderr, "Could not allocate space for structure stat[%d], aborting.\n", i);

		exit(1);
		}

	/* fill in structure stat */
	a = stat(input_filename[i], statptr[i]);
	if(a == -1)
		{
		fprintf(stderr, "stat filename[%d] failed, aborting.\n", i);
	
		exit(1);
		}

	if(debug_flag)
		{
		fprintf(stderr, "statptr[%d]=%p i=%d statptr[%d] -> st_size=%lu\n",\
		i, statptr[i], i, i, (unsigned long)statptr[i] -> st_size);
		}

	fprintf(stderr, "input file %s size is %lu delay is %.2f milli seconds\n",\
	input_filename[i], (unsigned long)statptr[i] -> st_size, delay[i]);

	/* get longest input file length */
	if( statptr[i] -> st_size > longest_input_file_length)
		{
		longest_input_file_length = statptr[i] -> st_size;
		}

	/* free structure stat */
	free(statptr[i]);

	} /* end for all input files */

if(debug_flag)
	{
	fprintf(stderr, "longest_input_file_length=%lu\n", (unsigned long)longest_input_file_length);
	}

/* create wave header for output file */
//char main_chunk[4];		/* 'RIFF' */
//uint32_t length;			/* length of file */
//char chunk_type[4];		/* 'WAVE' */
//char sub_chunk[4];		/* 'fmt' */
//uint32_t length_chunk;	/* length sub_chunk, always 16 bytes */
//uint16_t format;			/* always 1 = PCM-Code */
//uint16_t modus;			/* 1 = Mono, 2 = Stereo */
//uint32_t sample_fq;		/* Sample Freq */
//uint32_t byte_p_sec;		/* Data per sec */
//uint16_t byte_p_spl;		/* bytes per sample, 1=8 bit, 2=16 bit (mono) 2=8 bit, 4=16 bit (stereo) */
//uint16_t bit_p_spl;		/* bits per sample, 8, 12, 16 */
//char data_chunk[4];		/* 'data' */
//uint32_t data_length;		/* length of data */

for(i = 0; i < input_files; i++)
	{
	in_header[i] = (wave_header *) malloc(header_size);
	if(! in_header[i])
		{
		fprintf(stderr, "Could not allocate space for in_header[%d], aborting.\n", i);

		exit(1);
		}

	a = fread(in_header[i], sizeof(char), header_size, ifptr[i]);
	if(a < header_size)
		{
		fprintf(stderr, "Could not read wave header for file %s, aborting.\n", input_filename[i]);

		exit(1);
		}

	bytes_read[i] += header_size;
	}

strcpy(out_header -> main_chunk, "RIFF");
strcpy(out_header -> chunk_type, "WAVE"); 
strcpy(out_header -> sub_chunk, "fmt ");
strcpy(out_header -> data_chunk, "data");
out_header -> length_chunk = 16; //always 16
out_header -> format = 1; //PCM
out_header -> modus = input_files;  // 2; // stereo

/* calculate bytes corresponding to each file delay */
/* bytes per millisecond */
da = (double)in_header[0] -> byte_p_sec / 1000.0;

for(i = 0; i < input_files; i++)
	{
	/* bytes to shift in audio */
	db = da * (double)delay[i];

	bytes_to_add[i] = (size_t)db;
	a = bytes_to_add[i] % in_header[0] -> byte_p_spl;

	/* align on sample */
	if(a) bytes_to_add[i] += in_header[0] -> byte_p_spl - a;  

	if(debug_flag)
		{
		fprintf(stderr, "bytes_to_add[%d]=%lu\n", i, (unsigned long)bytes_to_add[i]);
		}

	/* so far nothing written */
	bytes_added[i] = 0;
	}

/* get longest input data length */
longest_input_data_length = 0;
for(i = 0; i < input_files; i++)
	{
	sta = in_header[i] -> data_length + bytes_to_add[i];
	if(sta > longest_input_data_length)
		{
		longest_input_data_length = sta;
		}
	}

if(debug_flag)
	{
	fprintf(stderr, "longest_input_data_length=%lu\n", (unsigned long)longest_input_data_length);
	}

out_header -> byte_p_sec = in_header[0] -> byte_p_sec * input_files;
out_header -> byte_p_spl = in_header[0] -> byte_p_spl; // 2 mono
out_header -> sample_fq = in_header[0] -> sample_fq;
out_header -> bit_p_spl = in_header[0] -> bit_p_spl; // 16
out_header -> data_length = longest_input_data_length * input_files;

strcpy(out_header -> data_chunk, "data");
out_header -> length = out_header -> data_length + header_size - 8;
if(debug_flag)
	{
	fprintf(stderr, "longest_input_file_length=%lu\n", (unsigned long)longest_input_file_length);
	fprintf(stderr, "out_header -> data_length=%lu\n", (unsigned long)out_header -> data_length);
	fprintf(stderr, "out_header -> length=%lu\n", (unsigned long)out_header -> length);
	}

expected_output_filesize = header_size + (input_files * longest_input_data_length);

fprintf(stderr, "expected_output_filesize is %lu\n", (unsigned long)expected_output_filesize);

fprintf(stderr, "writing wave format file\n");

bytes_written = 0;

/* write out wave header */
a = fwrite(out_header, sizeof(char), header_size, ofptr);
if(a != header_size)
	{
	fprintf(stderr,\
	"multimux(): could only write %d of %d bytes of out header, aborting.\n",\
	a, header_size);   

	exit(1); 
	}

bytes_written += a;

display_counter = 0;
while(1)
	{
	for(i = 0; i < input_files; i++)
		{
		/* write lead samples if delay specified for this channel */
		if(bytes_to_add[i] > bytes_added[i])
			{
			a = out_header -> byte_p_spl;
			for(j = 0; j < a; j++)
				{
				uc[j] = 0;
				}								

			bytes_added[i] += a;
			}
		else
			{
			a = fread(uc, sizeof(char), out_header -> byte_p_spl, ifptr[i]);
			}

		/* shorter files are padded with zeros */
		if(a < 1) /* EOF or error */
			{
			uc[0] = uc[1] = uc[2] = uc[3] = 0;
			}


		/* write the data */
		a = fwrite(uc, sizeof(char),  out_header -> byte_p_spl, ofptr);
		bytes_written += a;		
		display_counter += a;

		if(bytes_written == expected_output_filesize) break;

		} /* end for all possible input files */

	/* report progress every now end then... */
	if( ( display_counter > 1024) || (bytes_written == expected_output_filesize) )
		{
		da = 100.0 * ((double)bytes_written / (double)expected_output_filesize);

		fprintf(stderr,\
		"writing output file %s %lu (%.2f%%)\r", output_filename, (unsigned long)bytes_written, da);
		display_counter = 0;
		}

	if(bytes_written == expected_output_filesize) break;
	
	} /* end while */

/* close output file */
fclose(ofptr);

/* close input files */
for(i = 0; i < input_files; i++)
	{
	fclose(ifptr[i]);
	}

fprintf(stderr, "multimux ready\n");

exit(0);
} /* end function main */

